package de.kif.backend.util

import de.kif.backend.repository.*
import de.kif.backend.route.api.error
import de.kif.backend.route.api.success
import de.kif.common.*
import de.kif.common.model.Post
import io.ktor.application.call
import io.ktor.http.CacheControl
import io.ktor.http.ContentType
import io.ktor.http.HttpStatusCode
import io.ktor.response.cacheControl
import io.ktor.response.respondTextWriter
import io.ktor.routing.Route
import io.ktor.routing.get
import kotlinx.coroutines.channels.Channel
import kotlinx.coroutines.channels.broadcast
import kotlinx.coroutines.runBlocking
import kotlinx.html.currentTimeMillis
import kotlin.concurrent.thread
import kotlin.math.max

object PushService {

    internal var leastValidTimestamp = System.currentTimeMillis()
    var signature = Post.generateUrl()

    private val messages: MutableList<Pair<Long, Message>> = mutableListOf()

    private val channel = Channel<SSE>()
    val broadcast = channel.broadcast()

    /**
     * Save the message with the current timestamp
     */
    suspend fun notify(type: MessageType, repository: RepositoryType, id: Long) {
        val timestamp = System.currentTimeMillis()
        val message = Message(type, repository, id)

        synchronized(messages) {
            messages += Pair(timestamp, message)
        }

        channel.send(
            SSE.Message(
                MessageBox(
                    System.currentTimeMillis(),
                    signature,
                    true,
                    listOf(message)
                )
            )
        )
    }

    private fun getIndexOfTimestamp(timestamp: Long): Int {
        var index = messages.binarySearch(Pair(timestamp, Message.empty), compareBy { it.first })
        index = if (index < 0) {
            -index - 1
        } else {
            index
        }

        var element = messages.getOrNull(index - 1)
        while (element != null && element.first >= timestamp) {
            index--
            element = messages.getOrNull(index - 1)
        }

        return max(0, index)
    }

    /**
     * Get all messages created after the given timestamp.
     * The return Box has the current timestamp. If the leastValidTimestamp is less then the current timestamp set the
     * valid flag to false and return an empty message list.
     */
    fun getMessages(timestamp: Long): MessageBox {
        return if (timestamp < leastValidTimestamp) {
            MessageBox(
                System.currentTimeMillis(),
                signature,
                false,
                emptyList()
            )
        } else {
            val m = synchronized(messages) {
                val index = getIndexOfTimestamp(timestamp)
                messages.drop(index)
            }

            MessageBox(
                System.currentTimeMillis(),
                signature,
                true,
                m.map { it.second }
            )
        }
    }

    /**
     * Delete all messages before the given timestamp
     */
    fun gc(timestamp: Long) {
        leastValidTimestamp = timestamp

        synchronized(messages) {
            messages.removeIf {
                it.first < timestamp
            }
        }
    }

    fun ping() {
        runBlocking {
            channel.send(SSE.Ping)
        }
    }
}

sealed class SSE {
    class Message(val messageBox: MessageBox) : SSE()
    object Ping: SSE()
}

fun Route.pushService() {
    PushService.leastValidTimestamp = System.currentTimeMillis()

    get("/api/updates") {
        try {
            val timestamp = call.request.queryParameters["timestamp"]?.toLongOrNull() ?: 0

            val messageBox = PushService.getMessages(timestamp)

            call.response.cacheControl(CacheControl.NoCache(null))
            call.success(messageBox)
        } catch (_: Exception) {
            call.error(HttpStatusCode.InternalServerError)
        }
    }

    get("/api/events") {
        val events = PushService.broadcast.openSubscription()
        try {
            call.response.cacheControl(CacheControl.NoCache(null))
            call.respondTextWriter(contentType = ContentType.Text.EventStream) {
                write("retry: 1000\n")
                write("\n")

                for (event in events) {
                    write("id: 0\n")
                    when (event) {
                        is SSE.Ping -> {
                            write("event: ping\n")
                            write("data: ${PushService.signature}\n")
                        }
                        is SSE.Message -> {
                            write("event: update\n")
                            write("data: ${Serialization.stringify(MessageBox.serializer(), event.messageBox)}\n")
                        }
                    }
                    write("\n")
                    flush()
                }
            }
        } finally {
            events.cancel()
        }
    }

    RoomRepository.registerPushService()
    ScheduleRepository.registerPushService()
    TrackRepository.registerPushService()
    UserRepository.registerPushService()
    WorkGroupRepository.registerPushService()
    PostRepository.registerPushService()
    AnnouncementRepository.registerPushService()

    thread(
        start = true,
        isDaemon = true,
        name = "push-service"
    ) {
        var gc = currentTimeMillis() + 1000 * 60 * 10
        while (true) {
            Thread.sleep(500)
            PushService.ping()

            val now = currentTimeMillis()
            if (gc < now) {
                gc = now + 1000 * 60 * 10
                PushService.gc(now - 1000 * 60)
            }
        }

    }
}
